//
//  VMActivity.h
//  VMobSDK
//
//  Created by vmob.co on 5/29/13.
//  Copyright (c) 2013 vmob.co. All rights reserved.
//

#import "VMRequestManager.h"
#import "VMConstant.h"
#import <CoreLocation/CoreLocation.h>

/**
 `VMActivityManager` is a subclass of `VMRequestManager` provided to interact with VMob Activities.
 */
@interface VMActivityManager : VMRequestManager

/**
 Current activity stack in the `VMActivityManager` class.
 */

@property (nonatomic, readonly) NSMutableArray *stackActivity;

/**
 Timestamp for the last successful sent stack in the `VMActivityManager` class.
 */

@property (nonatomic, readonly) NSDate *lastSendStack;

/**
 Returns the shared activity manager object for the system.
 
 @return The systemwide activity manager.
 */

+ (instancetype) sharedInstance;

#pragma mark - Follow Methods
///--------------------------
/// @name Follow Methods
///--------------------------

/**
 Sets block called when sending the activity stack (returns NSDate).
 
 @param block block called when sending the activity stack. If nil is passed here, parameter will be nil.
 */

- (void) setUpdateDate: (ObjectBlock) block;

/**
 Sets block that is called when the activity stack is updated (returns NSArray).
 
 @param block block called when the activity stack is updated (returns NSArray). If nil is passed here, parameter will be nil.
 */

- (void) setUpdateStackActivity: (ObjectBlock) block;

#pragma mark - Activity API Methods

/**
 Method is provided to log an instance of a Client Application 'starting', e.g. coming out of hibernation or cold start.
 
 @param userToken Access token.
 @param completionBlock The block is executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block is executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */

- (VMHTTPRequestOperation *) sendHistoryActivityWithUserToken: (NSString *) userToken
											  completionBlock: (SuccessBlock) completionBlock
													failBlock: (ErrorBlock) errorBlock;
#pragma mark - Universal Methods

- (void) universalActivityWithActionTypeCode: (VMActionTypeCode) actionTypeCode
								  actionCode: (NSString *) actionCode
									  value1: (NSString *) value1
									  value2: (NSString *) value2
									  value3: (NSString *) value3
								redemptionId: (NSString *) redemptionId
									latitude: (NSString *) latitude
								   longitude: (NSString *) longitude
									 ownerId: (NSString *) ownerId
									   macId: (NSString *) macId
								  merchantId: (NSString *) merchantId
								  campaignId: (NSString *) campaignId
									  itemId: (NSString *) itemId
									itemCode: (NSString *) itemCode
									 venueId: (NSString *) venueId;

#pragma mark - ButtonClick Methods
///--------------------------
/// @name ButtonClick Methods
///--------------------------

/**
 This anonymous version for the Button Click service should only be used in situations where the user has not yet logged in or registered.

 @param buttonId should contain unique identifier for the button which was clicked
 @param merchantId should be supplied if relevant to the click
 @param venueId should be supplied if relevant to the click
 @param itemId should be supplied if relevant to the click
 @param itemCode identifies the type for the itemId, e.g. if itemId is an offer id then itemCode must be 'O'. itemCode is mandatory when itemId contains a value
 */

- (void) buttonClickAnonymousWithButtonId: (NSString *) buttonId
							   merchantId: (NSString *) merchantId
								  venueId: (NSString *) venueId
								   itemId: (NSString *) itemId
								 itemCode: (NSString *) itemCode;

/**
 The Button Click type is provided to log an instance of a Consumer clicking a button within the Client Application.
 
 @param buttonId should contain unique identifier for the button which was clicked
 @param merchantId should be supplied if relevant to the click
 @param venueId should be supplied if relevant to the click
 @param itemId should be supplied if relevant to the click
 @param itemCode identifies the type for the itemId, e.g. if itemId is an offer id then itemCode must be 'O'. itemCode is mandatory when itemId contains a value
 */

- (void) buttonClickWithButtonId: (NSString *) buttonId
					  merchantId: (NSString *) merchantId
						 venueId: (NSString *) venueId
						  itemId: (NSString *) itemId
						itemCode: (NSString *) itemCode;

#pragma mark - LocationCheckin Methods
///------------------------------
/// @name LocationCheckin Methods
///------------------------------

/**
 The Location Checkin type is provided to log the current location of the consumer/device
 
 @param typeGPS  this is where/how the GPS information was gathered from on the device
 @param location this is CLLocation
 */

- (void) locationCheckinWithTypeGPS: (VMTypeGPS) typeGPS
						   location: (CLLocation *) location;

#pragma mark - AppPageImpression Methods
///--------------------------------
/// @name AppPageImpression Methods
///--------------------------------

/**
 This anonymous version for the AppPageImpression service should only be used in situations where the user has not yet logged in or registered.
 
 @param pageId should contain unique identifier for the page which was viewed
 @param merchantId should be supplied if relevant to the current page
 @param venueId should be supplied if relevant to the current page
 @param itemId should be supplied if relevant to the current page
 @param itemCode identifies the type for the itemId, e.g. if itemId is an offer id then itemCode must be 'O'. itemCode is mandatory when itemId contains a value
 */

- (void) appPageImpressionAnonymWithPageId: (NSString *) pageId
								merchantId: (NSString *) merchantId
								   venueId: (NSString *) venueId
									itemId: (NSString *) itemId
								  itemCode: (NSString *) itemCode;

/**
 The AppPageImpression type is provided to log an instance of a Consumer viewing a page within the Client Application.
 
 @param pageId should contain unique identifier to the page which was viewed
 @param merchantId should be supplied if relevant to the current page
 @param venueId should be supplied if relevant to the current page
 @param itemId should be supplied if relevant to the current page
 @param itemCode identifies the type for the itemId, e.g. if itemId is an offer id then itemCode must be 'O'. itemCode is mandatory when itemId contains a value
 */

- (void) appPageImpressionWithPageId: (NSString *) pageId
						  merchantId: (NSString *) merchantId
							 venueId: (NSString *) venueId
							  itemId: (NSString *) itemId
							itemCode: (NSString *) itemCode;

#pragma mark - Offer Methods
///--------------------
/// @name Offer Methods
///--------------------

/**
 The OfferImpression type is provided to log an instance of an offer coming into the viewable area of the Client Application. 
 
 @param placementType should contain the placement type it originated from
 @param offerId should contain the offerId
 */

- (void) offerImpressionWithPlacementType: (NSString *) placementType
								  offerId: (NSString *) offerId;

/**
 The OfferRedemption type is provided to log an instance of a Consumer clicking to Redeem and Offer within the Client Application.
 
 @param placementType should contain the placement type it originated from
 @param paymentTypeCode this is Payment Type Code
 @param paymentId this is Payment Id if relevant
 @param paymentAmount this is Payment Amount if relevant
 @param redeemedOffer the call to create a redeemed offer
 @param offerId should contain the offerId
 @param venueId should be supplied if relevant to the current page
 */

- (void) offerRedemptionWithPlacementType: (NSString *) placementType
						  paymentTypeCode: (NSString *) paymentTypeCode
								paymentId: (NSString *) paymentId
							paymentAmount: (NSString *) paymentAmount
							redeemedOffer: (NSString *) redeemedOffer
								  offerId: (NSString *) offerId
								  venueId: (NSString *) venueId;

/**
 The OfferClickThrough type is provided to log an instance of a Consumer clicking on an offer to view the full detail within the Client Application.
 
 @param placementType should contain the placement type it originated from
 @param offerId should contain the offerId
 */

- (void) offerClickThroughWithPlacementType: (NSString *) placementType
									offerId: (NSString *) offerId;

/**
 The OfferInstoreRedemption typeis provided to log an instance of an 'in-store redemption' occuring.
 
 @param placementType should contain the placement type it originated from
 @param redeemedOffer the call to create a redeemed offer
 @param offerId should contain the offerId
 @param venueId should be supplied if relevant to the current page
 */

- (void) offerInstoreRedemptionWithPlacementType: (NSString *) placementType
								   redeemedOffer: (NSString *) redeemedOffer
										 offerId: (NSString *) offerId
										 venueId: (NSString *) venueId;

#pragma mark - Application Methods
///--------------------------
/// @name Application Methods
///--------------------------

/**
 The AppStartup type is provided to log an instance of a Client Application 'starting', e.g. coming out of hibernation or cold start. 
 */

- (void) appStartup;

/**
 This anonymous version of the AppStartup type should only be used in situations where the user has not yet registered or logged in.
 */

- (void) appStartupAnonym;

/**
 This type should only ever be posted once by the application - and at the earliest possible time, e.g. during installation oir first run (before user login)
 */

- (void) appInstall;

#pragma mark - Social Methods
///----------------------------
/// @name Social Methods
///----------------------------

/**
 The Registration type is provided to log an instance of a Consumer Registration occurring 
 
 @param socialType should contain the type of registration that has occurred
*/

- (void) registrationWithRegistrationType: (VMActivitySocialType) socialType;

/**
 The Login type is provided to log an instance of a Consumer Logging into the Client Application 
 
 @param socialType should contain the type of login that has occurred
 */

- (void) loginWithLoginType: (VMActivitySocialType) socialType;

/**
 The Offer Share type is provided to log an instance of a Consumer sharing an offer via a 3rd party application.
 
 @param socialType should contain the type of login that has occurred
 @param offerId is an offer
 */

- (void) offerShareWithSocialType: (VMShareSocialType) socialType
						  offerId: (NSString *) offerId;

#pragma mark - Advertisement Methods
///----------------------------
/// @name Advertisement Methods
///----------------------------

/**
 The Advertisement Impression type is provided to log an instance of an being displayed to the consumer on the app.
 
 @param placementCode should contain the placementCode for the Ad
 @param channelCode should contain the channelCode for the Ad
 @param advertisementId should be the id of the advertisement
 */

- (void) advertisementImpressionWithPlacementCode: (NSString *) placementCode
									  channelCode: (NSString *) channelCode
								  advertisementId: (NSString *) advertisementId;

/**
 The Advertisement Click type is provided to log an instance of a Consumer clicking on an advertisement within the application.
 
 @param placementCode should contain the placementCode for the Ad
 @param channelCode should contain the channelCode for the Ad
 @param advertisementId should be the id of the advertisement
 */

- (void) advertisementClickWithPlacementCode: (NSString *) placementCode
								 channelCode: (NSString *) channelCode
							 advertisementId: (NSString *) advertisementId;
@end
