//
//  VMConsumer.h
//  VMobSDK
//
//  Created by vmob.co on 5/30/13.
//  Copyright (c) 2013 vmob.co. All rights reserved.
//

#import "VMRequestManager.h"
#import "VMConsumer.h"
#import "VMConstant.h"

/**
 `VMConsumerManager` is a subclass of `VMRequestManager` designed for managing all interactions applicable to Consumers of the system.  This class is responsible for registering Consumers, logging the Consumer in or resetting Consumer's password.
*/
@interface VMConsumerManager : VMRequestManager

/**
 The latest Access Token received by the registration or login. Must be used in subsequent API calls.
 */
@property (nonatomic, readonly, strong) NSString *userToken;

/**
 Returns the shared consumer manager object for the system.
 
 @return The systemwide consumer manager.
 */
+ (instancetype) sharedInstance;

#pragma mark - Register Methods
///-----------------------
/// @name Register Methods
///-----------------------

/**
 The Create Mobile Phone Number Registration service is provided for registering of new Consumers using their mobile phone number as the Username. Calling the Create Mobile Phone Number Registration creates a new Consumer in the system, Logs the Consumer in and then creates and saves a valid OAuth Access Token. The saved OAuth Access Token can then be used in for subsequent API calls to identify the Consumer of the Client Application.
 
 @param mobilePhoneNumber Consumer's phone number. It must contain an international prefix (+61 ....)
 @param userName Consumer's user name.
 @param password Consumer's password. Password minimal length is 6 characters.
 @param firstName Consumer's first name. Either First and Lastname must be present OR the full name.
 @param lastName Consumer's last name. Either First and Lastname must be present OR the full name.
 @param fullName Consumer's full name. Either First and Lastname must be present OR the full name.
 @param gender Consumer's gender.
 @param dateOfBirth Consumer's date of birth.
 @param city Consumer's city.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) registerConsumerWithMobilePhone: (NSString *) mobilePhoneNumber
													userName: (NSString *) userName
													password: (NSString *) password
												   firstName: (NSString *) firstName
													lastName: (NSString *) lastName
													fullName: (NSString *) fullName
													  gender: (VMGender) gender
												 dateOfBirth: (NSDate *) dateOfBirth
														city: (NSString *) city
											 completionBlock: (ObjectBlock) completionBlock
												   failBlock: (ErrorBlock) errorBlock;

/**
 The Create Email Registration service is provided for registering of new Consumers using their email address as the Username. Calling the Create Email Registration creates a new Consumer in the system, Logs the Consumer in and then creates and saves a valid OAuth Access Token.  The saved OAuth Access Token can then be used in for subsequent API calls to identify the Consumer of the Client Application.
 
 @param email Consumer's email. The email address must be a valid email.
 @param userName Consumer's user name.
 @param password Consumer's password. Password minimal length is 6 characters.
 @param firstName Consumer's first name. Either First and Lastname must be present OR the full name
 @param lastName Consumer's last name. Either First and Lastname must be present OR the full name
 @param fullName Consumer's full name. Either First and Lastname must be present OR the full name
 @param gender Consumer's gender.
 @param dateOfBirth Consumer's date of birth.
 @param city Consumer's city. 
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) registerConsumerWithEmail: (NSString *) email
											  userName: (NSString *) userName
											  password: (NSString *) password
											 firstName: (NSString *) firstName
											  lastName: (NSString *) lastName
											  fullName: (NSString *) fullName
												gender: (VMGender) gender
										   dateOfBirth: (NSDate *) dateOfBirth
												  city: (NSString *) city
									   completionBlock: (ObjectBlock) completionBlock
											 failBlock: (ErrorBlock) errorBlock;

/**
 The Create Anonymous Registration service is provided for client application to register without providing any Consumer details. A unique username and password is still required as the system will create an unique Anonymous Consumer record. This username and password can be discarded, only the saved Access Token is required to make subsequent API calls.
 
 @param userName Consumer's user name.
 @param password Consumer's password. Password minimal length is 6 characters.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) registerConsumerWithAnonymousUserName: (NSString *) userName
														  password: (NSString *) password
												   completionBlock: (ObjectBlock) completionBlock
														 failBlock: (ErrorBlock) errorBlock;

#pragma mark - Login Methods
///--------------------
/// @name Login Methods
///--------------------

/**
 The Login with Username service is provided for existing Consumers to login using their username and password. The username will be either an email address or mobile phone number depending on how the Consumer had initially registered.  Login with Username will typically be used when logging into a second or subsequent Device, or via the browser etc. This is because in the normal process flow, a Consumer will register and the Access Token will be returned and cached. The Access Token is then must be used when starting the application instead of getting the Consumer to log in again.
 Calling the Login with Username logs the Consumer into the system and creates and saves a valid OAuth Access Token.  The saved OAuth Access Token can then be used in for subsequent API calls to identify the Consumer of the Client Application.
 
 @param userName Consumer's user name. Username and password are required.
 @param password Consumer's password. Username and password are required.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) loginConsumerWithUserName: (NSString *) userName
											  password: (NSString *) password
									   completionBlock: (ObjectBlock) completionBlock
											 failBlock: (ErrorBlock) errorBlock;

/**
 The Login with Social service is provided for either new or existing Consumers to login using 3rd party credentials. (e.g. Facebook / Twitter).  The API will resolve the 'Username' to a genuine 3rd party Access Token (e.g. in the case of Facebook it will just be a code which can then be used to retrieve an access token) and then retrieve basic user information, before returning a 'VMob' Access Token. The `socialType` defines which 3rd Party site and authentication method to use, e.g. `VMFacebook`.
 Calling the Login with Social logs the Consumer into the system and creates and saves a valid OAuth Access Token. The saved OAuth Access Token can then be used in for subsequent API calls to identify the Consumer of the Client Application.
 
 @param token 3rd party token.
 @param socialType Defines which 3rd Party site and authentication method to use.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) loginConsumerWithSocialToken: (NSString *) token
											   socialType: (VMSocialType) socialType
										  completionBlock: (ObjectBlock) completionBlock
												failBlock: (ErrorBlock) errorBlock;
#pragma mark - Detail Methods
///---------------------
/// @name Detail Methods
///---------------------

/**
 Gets the current consumer details.
 
 @param token Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) detailConsumerWithUserToken: (NSString *) token
										 completionBlock: (ObjectBlock) completionBlock
											   failBlock: (ErrorBlock) errorBlock;

/**
 Updates a subset of consumer properties.
 
 @param token Access token.
 @param consumer Updated `VMConsumer` object.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) updateConsumerWithUserToken: (NSString *) token
												consumer: (VMConsumer *) consumer
										 completionBlock: (ObjectBlock) completionBlock
											   failBlock: (ErrorBlock) errorBlock;

/**
 Updates a subset of consumer properties.
 
 @param token Access token.
 @param firstName Consumer's first name. Either both the `firstName` and `lastName` properties or just the `fullName` must be supplied, it is not necessary to supply both.
 @param lastName Consumer's last name. Either both the `firstName` and `lastName` properties or just the `fullName` must be supplied, it is not necessary to supply both.
 @param fullName Consumer's full name. Either both the `firstName` and `lastName` properties or just the `fullName` must be supplied, it is not necessary to supply both.
 @param gender Consumer's gender.
 @param dateOfBirth Consumer's date of birth.
 @param city Consumer's city. Home City is a new property, if not implemented on the client application yet, please specify as '2'
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 @param emailAddress email Address update
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) updateConsumerWithUserToken: (NSString *) token
											   firstName: (NSString *) firstName
												lastName: (NSString *) lastName
												fullName: (NSString *) fullName
												  gender: (VMGender) gender
											 dateOfBirth: (NSDate *) dateOfBirth
													city: (NSString *) city
                                            emailAddress: (NSString *) emailAddress
										 completionBlock: (ObjectBlock) completionBlock
											   failBlock: (ErrorBlock) errorBlock;

#pragma mark - RedeemOffers Methods
///---------------------------
/// @name RedeemOffers Methods
///---------------------------

/**
 Takes a snapshot of the current state of the offer and creates a 'redeemed offer' resource.
 
 @param token Access token.
 @param offerId offerID.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) createRedeemedOffersWithUserToken: (NSString *) token
													   offerId: (NSString *) offerId
											   completionBlock: (ObjectBlock) completionBlock
													 failBlock: (ErrorBlock) errorBlock;

/**
 Gets a list of offers which have been redeemed by the consumer
 
 @param token Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) listRedeemedOffersWithUserToken: (NSString *) token
											 completionBlock: (ObjectBlock) completionBlock
												   failBlock: (ErrorBlock) errorBlock;

/**
 Sets a redeemed offer's status to hidden.
 
 @param token Access token.
 @param offerId offerID.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) deleteRedeemedOffersWithUserToken: (NSString *) token
													   offerId: (NSString *) offerId
											   completionBlock: (ObjectBlock) completionBlock
													 failBlock: (ErrorBlock) errorBlock;
#pragma mark - Tags Methods
///---------------------------
/// @name Tags Methods
///---------------------------

/**
 Gets the list of the tag values that are assigned to the current consumer. The string value returned for each tag is the referenceCode of tag value (unique text code).
 
 @param token Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) listTagWithUserToken: (NSString *) token
								  completionBlock: (ObjectBlock) completionBlock
										failBlock: (ErrorBlock) errorBlock;
/**
 Attempts to add a range of tag values to a consumer. Any invalid or duplicate values will fail silently.
 
 @param token Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 @param tagValueReferenceCodes tags Value Reference Codes being added.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) addTagsWithUserToken: (NSString *) token
								  completionBlock: (ObjectBlock) completionBlock
										failBlock: (ErrorBlock) errorBlock
						   tagValueReferenceCodes: (NSString *) tagValueReferenceCode, ... NS_REQUIRES_NIL_TERMINATION;
/**
 Attempts to remove a range of tag values from the consumer. Any invalid or missing values will fail silently.
 
 @param token Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 @param tagValueReferenceCodes remove tags Value Reference Codes.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) removeTagsWithUserToken: (NSString *) token
									 completionBlock: (ObjectBlock) completionBlock
										   failBlock: (ErrorBlock) errorBlock
							  tagValueReferenceCodes: (NSString *) tagValueReferenceCode, ... NS_REQUIRES_NIL_TERMINATION;

/**
 Attempts to add a range of tag values to a consumer and remover a range of tag values from a consumer. Any invalid or duplicate values will fail silently.
 
 @param token Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 @param tagValueAddReferenceCodes added tags Value Reference Codes.
 @param tagValueRemoveReferenceCodes remove tags Value Reference Codes.
 
 @return A new request operation.
 */

- (VMHTTPRequestOperation *) updateTagWithUserToken: (NSString *) token
						  tagValueAddReferenceCodes: (NSArray *) tagValueAddReferenceCodes
					   tagValueRemoveReferenceCodes: (NSArray *) tagValueRemoveReferenceCodes
									completionBlock: (ObjectBlock) completionBlock
										  failBlock: (ErrorBlock) errorBlock;

#pragma mark - CrossReference Methods
///---------------------------
/// @name CrossReference Methods
///---------------------------

/**
Cross References are created between the core data and 3rd parties such as GCM. For example, when a device is registered with GCM, it sends the registration through this api to create the cross reference between consumer and GCM
 
 @param token Push Notification token.
 @param userToken Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */

- (VMHTTPRequestOperation *) createCrossReferenceWithPushNotification: (NSString *) token
															userToken: (NSString *) userToken
													  completionBlock: (ObjectBlock) completionBlock
															failBlock: (ErrorBlock) errorBlock;
/**
 Cross References are created between the core data and 3rd parties such as GCM. For example, when a device is registered with GCM, it sends the registration through this api to create the cross reference between consumer and gcm
 
 @param externalId external Id
 @param token is the ID supplied by the 3rd party system.
 @param userToken Access token.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) createCrossReferenceWithExternalId: (NSString *) externalId
													  userToken: (NSString *) userToken
													 systemType: (VMCrossReferenceType) crossReferenceType
												completionBlock: (ObjectBlock) completionBlock
													  failBlock: (ErrorBlock) errorBlock;
@end
