//
//  VMOfferManager.h
//  VMobSDK
//
//  Created by vmob.co on 6/1/13.
//  Copyright (c) 2013 vmob.co. All rights reserved.
//

#import "VMRequestManager.h"

/**
 `VMOfferManager` is a subclass of `VMRequestManager` providing all 'offer' related information available to a consumer.
 */
@interface VMOfferManager : VMRequestManager

/**
 Returns the shared offer manager object for the system.
 
 @return The systemwide offer manager.
 */
+ (instancetype) sharedInstance;

/**
 Gets a list of merchants who have active offers, optionally filtered by a category id.
 
 @param token Access token.
 @param categoryId Category ID.
 @param offset Offset.
 @param limit Limit.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) merchantListWithUserToken: (NSString *) token
											categoryId: (NSString *) categoryId
												offset: (NSString *) offset
												 limit: (NSString *) limit
									   completionBlock: (ObjectBlock) completionBlock
											 failBlock: (ErrorBlock) errorBlock;

/**
 Gets the full list of main / top-level categories, boolean parameter specifies whether or not to include categories which the current consumer has actively disabled.
 
 @param token Access token.
 @param includeConsumerExcludedCategories Parameter specifies whether or not to include categories which the current consumer has actively disabled.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) categoriesOfferWithUserToken: (NSString *) token
						includeConsumerExcludedCategories: (BOOL) includeConsumerExcludedCategories
										  completionBlock: (ObjectBlock) completionBlock
												failBlock: (ErrorBlock) errorBlock;

/**
 Offers are returned in the order of those that have the closest venue come first, the details and distance (in meters) to the closest venue are also returned.
 
 @param token Access token.
 @param lat Latitude.
 @param lon Longitude.
 @param categoryId Category ID. CategoryId is an optional filter that can be supplied, pass an empty or nil value to search all categories.
 @param offset Offset.
 @param limit Limit.
 @param fields Specify exact field names as a comma separated list in the `fields` parameter to restrict the population of fields with data.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) nearbyOffersWithUserToken: (NSString *) token
											  latitude: (NSString *) lat
											 longitude: (NSString *) lon
											categoryId: (NSString *) categoryId
												offset: (NSString *) offset
												 limit: (NSString *) limit
												fields: (NSString *) fields
									   completionBlock: (ObjectBlock) completionBlock
											 failBlock: (ErrorBlock) errorBlock;

/**
 Offers are returned by one of 2 formats, either 'Ranked Offers' which have been ranked by a special criteria, e.g. 'Hot Deals' OR queried by category, merchant and/or keyword.
 
 @param token Access token.
 @param ranked Ranked.
 @param keyword Keyword.
 @param categoryId Category ID. CategoryId is an optional filter that can be supplied, pass an empty or nil value to search all categories.
 @param merchantId merchantId
 @param fields Specify exact field names as a comma separated list in the `fields` parameter to restrict the population of fields with data.
 @param orderBy OrderBy is case-sensitive and defaults to `kVMOrderByWeighting`, other options are `kVMOrderByLastUpdatedAt` and `kVMOrderByStartDate`
 @param ascending Order direction options are `YES` for Ascending and `NO` for Descending.
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
 
 @return A new request operation.
 */
- (VMHTTPRequestOperation *) searchOffersWithUserToken: (NSString *) token
										  rankedSearch: (BOOL) ranked
											   keyword: (NSString *) keyword
											categoryId: (NSString *) categoryId
											merchantId: (NSString *) merchantId
												fields: (NSString *) fields
											   orderBy: (NSString *) orderBy
											 ascending: (BOOL) ascending
									   completionBlock: (ObjectBlock) completionBlock
											 failBlock: (ErrorBlock) errorBlock;

/**
 Gets the terms and conditions for a given offer id.
 
 @param token Access token.
 @param offerId Offer Id
 @param completionBlock The block to be executed on the completion of a successful request. This block has no return value and takes one argument: the object constructed from the response data of the request.
 @param errorBlock The block to be executed on the completion of an unsuccessful request. This block has no return value and takes one argument: the error that occurred during the request.
*/

- (VMHTTPRequestOperation *) termsAndConditionsWithUserToken: (NSString *) token
                                                     offerId: (NSString *) offerId
                                             completionBlock: (ObjectBlock) completionBlock
                                                   failBlock: (ErrorBlock) errorBlock;

@end
