// UIImageView+VMNetworking.m
//
// Copyright (c) 2011 Gowalla (http://gowalla.com/)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#import <Foundation/Foundation.h>
#import <objc/runtime.h>

#if __IPHONE_OS_VERSION_MIN_REQUIRED
#import "UIImageView+VMNetworking.h"

@interface VMImageCache : NSCache
- (UIImage *)cachedImageForRequest:(NSURLRequest *)request;
- (void)cacheImage:(UIImage *)image
        forRequest:(NSURLRequest *)request;
@end

#pragma mark -

static char kVMImageRequestOperationObjectKey;

@interface UIImageView (_VMNetworking)
@property (readwrite, nonatomic, strong, setter = VM_setImageRequestOperation:) VMImageRequestOperation *VM_imageRequestOperation;
@end

@implementation UIImageView (_VMNetworking)
@dynamic VM_imageRequestOperation;
@end

#pragma mark -

@implementation UIImageView (VMNetworking)

- (VMHTTPRequestOperation *)VM_imageRequestOperation {
    return (VMHTTPRequestOperation *)objc_getAssociatedObject(self, &kVMImageRequestOperationObjectKey);
}

- (void)VM_setImageRequestOperation:(VMImageRequestOperation *)imageRequestOperation {
    objc_setAssociatedObject(self, &kVMImageRequestOperationObjectKey, imageRequestOperation, OBJC_ASSOCIATION_RETAIN_NONATOMIC);
}

+ (NSOperationQueue *)VM_sharedImageRequestOperationQueue {
    static NSOperationQueue *_VM_imageRequestOperationQueue = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        _VM_imageRequestOperationQueue = [[NSOperationQueue alloc] init];
        [_VM_imageRequestOperationQueue setMaxConcurrentOperationCount:NSOperationQueueDefaultMaxConcurrentOperationCount];
    });
    
    return _VM_imageRequestOperationQueue;
}

+ (VMImageCache *)VM_sharedImageCache {
    static VMImageCache *_VM_imageCache = nil;
    static dispatch_once_t oncePredicate;
    dispatch_once(&oncePredicate, ^{
        _VM_imageCache = [[VMImageCache alloc] init];
    });
    
    return _VM_imageCache;
}

#pragma mark -

- (void)setImageWithURL:(NSURL *)url {
    [self setImageWithURL:url placeholderImage:nil];
}

- (void)setImageWithURL:(NSURL *)url 
       placeholderImage:(UIImage *)placeholderImage
{
    NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url];
    [request setHTTPShouldHandleCookies:NO];
    [request addValue:@"image/*" forHTTPHeaderField:@"Accept"];
    
    [self setImageWithURLRequest:request placeholderImage:placeholderImage success:nil failure:nil];
}

- (void)setImageWithURLRequest:(NSURLRequest *)urlRequest 
              placeholderImage:(UIImage *)placeholderImage 
                       success:(void (^)(NSURLRequest *request, NSHTTPURLResponse *response, UIImage *image))success
                       failure:(void (^)(NSURLRequest *request, NSHTTPURLResponse *response, NSError *error))failure
{
    [self cancelImageRequestOperation];
    
    UIImage *cachedImage = [[[self class] VM_sharedImageCache] cachedImageForRequest:urlRequest];
    if (cachedImage) {
        self.image = cachedImage;
        self.VM_imageRequestOperation = nil;
        
        if (success) {
            success(nil, nil, cachedImage);
        }
    } else {
        self.image = placeholderImage;
        
        VMImageRequestOperation *requestOperation = [[VMImageRequestOperation alloc] initWithRequest:urlRequest];
        [requestOperation setCompletionBlockWithSuccess:^(VMHTTPRequestOperation *operation, id responseObject) {
            if ([[urlRequest URL] isEqual:[[self.VM_imageRequestOperation request] URL]]) {
                if (success) {
                    success(operation.request, operation.response, responseObject);
                } else {
                    self.image = responseObject;
                }
                
                self.VM_imageRequestOperation = nil;
            }

            [[[self class] VM_sharedImageCache] cacheImage:responseObject forRequest:urlRequest];
        } failure:^(VMHTTPRequestOperation *operation, NSError *error) {
            if ([[urlRequest URL] isEqual:[[self.VM_imageRequestOperation request] URL]]) {
                if (failure) {
                    failure(operation.request, operation.response, error);
                }
                
                self.VM_imageRequestOperation = nil;
            }
        }];
        
        self.VM_imageRequestOperation = requestOperation;
        
        [[[self class] VM_sharedImageRequestOperationQueue] addOperation:self.VM_imageRequestOperation];
    }
}

- (void)cancelImageRequestOperation {
    [self.VM_imageRequestOperation cancel];
    self.VM_imageRequestOperation = nil;
}

@end

#pragma mark -

static inline NSString * VMImageCacheKeyFromURLRequest(NSURLRequest *request) {
    return [[request URL] absoluteString];
}

@implementation VMImageCache

- (UIImage *)cachedImageForRequest:(NSURLRequest *)request {
    switch ([request cachePolicy]) {
        case NSURLRequestReloadIgnoringCacheData:
        case NSURLRequestReloadIgnoringLocalAndRemoteCacheData:
            return nil;
        default:
            break;
    }
    
	return [self objectForKey:VMImageCacheKeyFromURLRequest(request)];
}

- (void)cacheImage:(UIImage *)image
        forRequest:(NSURLRequest *)request
{
    if (image && request) {
        [self setObject:image forKey:VMImageCacheKeyFromURLRequest(request)];
    }
}

@end

#endif
