//
//  ServiceProvider.m
//
//  Created by Anton Kaizer on 04.02.13.
//  Copyright (c) 2013 Arello Mobile. All rights reserved.
//

#import "VMHTTPProvider.h"
#import "VMJSRequestOperation.h"
#import "NSString+VMValidation.h"
#import "NSDictionary+VMValidation.h"
#import "NSString+VMToolkitAdditions.h"

#define kAMMaxRequestBodyLengthToLogging 1024

@interface VMHTTPProvider () 

@end

@implementation VMHTTPProvider

#pragma mark singleton

+ (instancetype) sharedProvider {
	static dispatch_once_t once;
    static id sharedInstance;
    dispatch_once(&once, ^{
        sharedInstance = [[self alloc] init];
    });
    return sharedInstance;
}

#pragma mark public

- (id) initWithBaseURL:(NSURL *)url {
	if (self = [super initWithBaseURL:url]) {
		self.parameterEncoding = VMJSONParameterEncoding;
		_responseKey = @"response";
		
#ifdef DEBUG
		self.enableLogging = YES;
#else
		self.enableLogging = NO;
#endif
		[self registerOperationClasses];
	}
	return self;
}

- (NSMutableURLRequest *) requestWithMethod:(NSString *)method path:(NSString *)path parameters:(NSDictionary *)parameters {
	NSMutableURLRequest *request = nil;
	
	if ([self.requestEnvelope isValid] && !([method isEqualToString:@"GET"] || [method isEqualToString:@"HEAD"])) { //|| [method isEqualToString:@"DELETE"])) {
		request = [super requestWithMethod:method path:path parameters:nil];
		NSString *charset = (__bridge NSString *)CFStringConvertEncodingToIANACharSetName(CFStringConvertNSStringEncodingToEncoding(self.stringEncoding));
		NSError *error = nil;
		NSString *requestString = [[NSString alloc] initWithData:[NSJSONSerialization dataWithJSONObject:parameters options:0 error:&error] encoding:self.stringEncoding];
		requestString = [NSString stringWithFormat:@"%@=%@", self.requestEnvelope, requestString];
		[request addValue:[NSString stringWithFormat:@"application/x-www-form-urlencoded; charset=%@", charset] forHTTPHeaderField:@"Content-Type"];
		[request setHTTPBody:[requestString dataUsingEncoding:self.stringEncoding]];
	} else {
		request = [super requestWithMethod:method path:path parameters:parameters];
	}
	return request;
}

-  (VMHTTPRequestOperation *) HTTPRequestOperationWithRequest:(NSMutableURLRequest *) request serviceRequest:(VMServiceRequest *) serviceRequest success:(RequestSuccessCallback) success failure:(RequestFailCallback) failure {
	
	if ([serviceRequest acceptableContentType])
		[request setValue:[serviceRequest acceptableContentType] forHTTPHeaderField:@"Accept"];
	
	[self requestWillSend:request];
	
	[self log:@"request url:%@", request.URL];
	[self log:@"request headers:%@", request.allHTTPHeaderFields];
	if (request.HTTPBody.length > 0 && request.HTTPBody.length < kAMMaxRequestBodyLengthToLogging)
		[self log:@"request body: %@", [[NSString alloc] initWithData:request.HTTPBody encoding:self.stringEncoding]];
	
	VMHTTPRequestOperation *operation = [self HTTPRequestOperationWithRequest:request success:^(VMHTTPRequestOperation *operation, id responseObject) {
		[self log:@"Response code: %d", operation.response.statusCode];
		[self log:@"Response headers: %@", operation.response.allHeaderFields];
		if (([operation.response.MIMEType hasPrefix:@"application"] || [operation.response.MIMEType hasPrefix:@"text"]) && operation.responseData.length < kAMMaxRequestBodyLengthToLogging) {
			[self log:@"Response body: %@", [[NSString alloc] initWithData:operation.responseData encoding:self.stringEncoding]];
		}
		
		VMRequestError *error = [serviceRequest validateResponse:responseObject httpResponse:operation.response];
		if (error) {
			[self log:@"request error:%@", error];
			if (failure)
				failure(serviceRequest, error);
		} else {
			dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
				NSObject *validResponseObject = responseObject;
//				if ([_responseKey isValid] && [validResponseObject isKindOfClass:[NSDictionary class]]) {
//					validResponseObject = [(NSDictionary*)responseObject objectForKey:_responseKey];
//				}
				[serviceRequest processResponse:validResponseObject];
				dispatch_async(dispatch_get_main_queue(), ^{
					if (success)
						success(serviceRequest);
				});
			});
		}
	} failure:^(VMHTTPRequestOperation *operation, NSError *error) {
		[self log:@"Response code: %d", operation.response.statusCode];
		[self log:@"Response headers: %@", operation.response.allHeaderFields];
		if (([operation.response.MIMEType hasPrefix:@"application"] || [operation.response.MIMEType hasPrefix:@"text"]) && operation.responseData.length < kAMMaxRequestBodyLengthToLogging) {
			[self log:@"Response body: %@", [[NSString alloc] initWithData:operation.responseData encoding:self.stringEncoding]];
		}
		[self log:@"request error: %@", error];
		
		if (failure)
			failure(serviceRequest, error);
	}];
	
	return operation;
}

- (VMHTTPRequestOperation *) sendRequest:(VMServiceRequest *) serviceRequest toPath:(NSString *) path success:(RequestSuccessCallback) success failure:(RequestFailCallback) failure {
	NSDictionary *requestDict = [serviceRequest requestDictionary];
	
	NSMutableURLRequest *request = [self requestWithMethod:[serviceRequest method] path:path parameters:requestDict];
	
	VMHTTPRequestOperation *operation = [self HTTPRequestOperationWithRequest:request serviceRequest:serviceRequest success:success failure:failure];
	[self enqueueHTTPRequestOperation:operation];
	return operation;
}

- (VMHTTPRequestOperation *) sendRequest:(VMServiceRequest *) serviceRequest success:(RequestSuccessCallback) success failure:(RequestFailCallback) failure {
	return [self sendRequest:serviceRequest toPath:serviceRequest.path success:success failure:failure];
}

#pragma mark private

- (void) registerOperationClasses {
	[self registerHTTPOperationClass: [VMJSRequestOperation class]];
}

- (void) requestWillSend:(NSMutableURLRequest *) request {
	NSString *requestId = [NSString stringWithUUID];
	[request setValue:requestId forHTTPHeaderField:@"X-Request-id"];
	if (self.autorizationToken)
		[request setValue:self.autorizationToken forHTTPHeaderField:@"X-Auth-Token"];
	[self log:@"sending request with id:%@", requestId];
}

- (VMHTTPRequestOperation *) uploadFileWithRequest:(VMFileRequest *) fileRequest success:(RequestSuccessCallback) success failure:(RequestFailCallback) failure  {
	
	NSMutableURLRequest *request = [self multipartFormRequestWithMethod:@"POST" path:fileRequest.path parameters:fileRequest.requestDictionary constructingBodyWithBlock:^(id<VMMultipartFormData> formData) {
		NSError *error = nil;
		if (fileRequest.filePath) {
			[formData appendPartWithFileURL:[NSURL fileURLWithPath:fileRequest.filePath] name:@"data[file]" error:&error];
		} else if (fileRequest.fileData) {
			[formData appendPartWithFileData:fileRequest.fileData name:@"data[file]" fileName:fileRequest.fileName mimeType:@"application/octet-stream"];
		}
		if (error) {
			failure(fileRequest, error);
		}
	}];
	
	VMHTTPRequestOperation *operation = [self HTTPRequestOperationWithRequest:request serviceRequest:fileRequest success:success failure:failure];
	
	[self enqueueHTTPRequestOperation:operation];
	
	return operation;
}

- (VMHTTPRequestOperation *) downloadFileWithRequest:(VMFileRequest *) fileRequest success:(RequestSuccessCallback) success failure:(RequestFailCallback) failure  {
	
	NSMutableURLRequest *request = [self requestWithMethod:fileRequest.method path:fileRequest.path parameters:fileRequest.requestDictionary];

	[self requestWillSend:request];
	
	VMHTTPRequestOperation *operation = [self HTTPRequestOperationWithRequest:request serviceRequest:fileRequest success:success failure:failure];
	
	if ([fileRequest filePath]) {
		operation.outputStream = [NSOutputStream outputStreamToFileAtPath:[fileRequest filePath] append:NO];
	}
	
	[operation setCompletionBlockWithSuccess:^(VMHTTPRequestOperation *operation, id responseObject) {
		[self log:@"Response code: %d", operation.response.statusCode];
		[self log:@"Response headers: %@", operation.response.allHeaderFields];
		
		if (operation.response.statusCode == HTTP_OK || operation.response.statusCode == HTTP_CREATED || operation.response.statusCode == HTTP_ACCEPTED || operation.response.statusCode == HTTP_NO_CONTENT) {
			if (![fileRequest filePath]) {
				fileRequest.fileData = operation.responseData;
				fileRequest.fileName = [fileRequest.path lastPathComponent];
			}
			if (success)
				success(fileRequest);
		} else {
			failure(fileRequest, [VMRequestError requestErrorWithCode:operation.response.statusCode description:@"Some server error occured."]);
		}
	} failure:^(VMHTTPRequestOperation *operation, NSError *error) {
		[self log:@"request error:%@", error];
		failure(fileRequest, error);
	}];
    [self enqueueHTTPRequestOperation:operation];
	
	return operation;
}

- (void) log:(NSString *) logFormat, ... {
	if (self.enableLogging) {
		va_list argumentList;
		va_start(argumentList, logFormat);
			
		NSLogv(logFormat, argumentList);
			
		va_end(argumentList);
	}
}

#pragma mark -

@end
