//
//  VMActivity.m
//  VMobSDK
//
//  Created by vmob.co on 5/29/13.
//  Copyright (c) 2013 vmob.co. All rights reserved.
//

#import "VMActivityManager.h"
#import "VMUtils.h"
#import "VMActivitiesRequest.h"
#import "VMConsumerManager.h"
#import "VMActivity.h"
#import "VMLog.h"

#define kSaveDate @"SaveDate"

@interface VMActivityManager()
@property (nonatomic, strong) NSMutableArray *sentStackActivity;
@property (nonatomic, strong) NSTimer *sendTimer;
@property (nonatomic) NSUInteger sizeStack;
@property (nonatomic) NSTimeInterval sendInterval;
@property (nonatomic) UIBackgroundTaskIdentifier backgroundSendID;
@property (nonatomic, copy) ObjectBlock updateDateBlock;
@property (nonatomic, copy) ObjectBlock updateStackActivityBlock;
@end;

@implementation VMActivityManager

- (id)init {
	if (self = [super init]) {
		_sizeStack = [VMUtils activitiesBatchSize];
		_sendInterval = [VMUtils activitiesSendMinTimerInSeconds];
		serviceProvider = [[VMServiceProvider alloc] initWithBaseURL:[NSURL URLWithString:[VMUtils activityApiUrl]]];
		_stackActivity = [NSMutableArray arrayWithContentsOfFile:[self stackActivityPatch]];
		[self validateActivitiesLog];
		self.sentStackActivity = [NSMutableArray array];
		if (!_stackActivity) {
			_stackActivity = [NSMutableArray new];
		}
		else {
			[self sendCurrentStackActivities];
		}
		
		//added Notification
		self.backgroundSendID = UIBackgroundTaskInvalid;
		[[NSNotificationCenter defaultCenter] addObserver:self
												 selector:@selector(willResignActive:)
													 name:UIApplicationWillResignActiveNotification
												   object:nil];
		[[NSNotificationCenter defaultCenter] addObserver:self
												 selector:@selector(didEnterForeground:)
													 name:UIApplicationWillEnterForegroundNotification
												   object:nil];
	}
	return self;
}

+ (instancetype) sharedInstance {
	static dispatch_once_t once;
    static id sharedInstance;
    dispatch_once(&once, ^{
        sharedInstance = [[self alloc] init];
    });
    return sharedInstance;
}

#pragma mark - Validation Methods

- (void) validateActivitiesLog {
	if (![_stackActivity count]) {
		return;
	}
	NSArray *array = [_stackActivity sortedArrayUsingComparator:^NSComparisonResult(NSDictionary *obj1, NSDictionary *obj2) {
		return ![VMActivity compareFirstDict:obj1 secondDict:obj2];
	}];
	
	if ([array count] > [VMUtils activitiesExpireSize]) {
		array = [array subarrayWithRange:NSMakeRange([array count] - [VMUtils activitiesExpireSize], [VMUtils activitiesExpireSize])];
	}
	_stackActivity = [NSMutableArray arrayWithArray:array];
	for(NSDictionary *dict in array) {
		if([VMActivity validateActivityWithDictionary:dict]) {
			break;
		}
		else {
			[_stackActivity removeObject:dict];
		}
	}
}

#pragma mark - Follow Methods

- (void) setUpdateDate: (ObjectBlock) block {
	self.updateDateBlock = block;
}

- (void) setUpdateStackActivity: (ObjectBlock) block {
	self.updateStackActivityBlock = block;
}

#pragma mark - Private Methods

- (NSString *) stackActivityPatch {
	NSString *documentDirectories = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) lastObject];
	return [NSString stringWithFormat:@"%@/%@", documentDirectories, @"stackActivity"];
}

- (void) sendLastStack {
	NSDate *date = [NSDate date];
	if(self.updateDateBlock) self.updateDateBlock(date);
	[[NSUserDefaults standardUserDefaults] setObject:date forKey:kSaveDate];
	[[NSUserDefaults standardUserDefaults] synchronize];
}

- (NSDate *) lastSendStack {
	return [[NSUserDefaults standardUserDefaults] objectForKey:kSaveDate];
}

#pragma mark - DataControl Methods

- (NSArray *) getSendingActivity {
	NSUInteger sendCount = _sizeStack > [_stackActivity count] ? [_stackActivity count]:_sizeStack;
	NSArray *sentCurrentStack = [_stackActivity subarrayWithRange:NSMakeRange([_stackActivity count] - sendCount, sendCount)];
	[_sentStackActivity addObjectsFromArray:sentCurrentStack];
	[_stackActivity removeObjectsInArray:sentCurrentStack];
	
	if(self.updateStackActivityBlock) self.updateStackActivityBlock(_stackActivity);
	return sentCurrentStack;
}

- (BOOL) checkSendingStack {
	if ([_stackActivity count] >= _sizeStack) {
		[self sendCurrentStackActivities];
		return YES;
	}
	return NO;
}


- (void) startTimer {
	if (!self.sendTimer && serviceProvider.networkReachabilityStatus > 0) {
		self.sendTimer = [NSTimer scheduledTimerWithTimeInterval:_sendInterval target:self selector:@selector(sendCurrentStackActivities) userInfo:nil repeats:YES];
	}
}

- (void) addedPullActivityDictionary: (NSDictionary *) dictionary {
#if DEBUG
	[VMLog log:[NSString stringWithFormat:@"ActivityDictionary -%@",dictionary]];
#endif
	[self startTimer];
	[_stackActivity addObject:dictionary];
	[self checkSendingStack];
	if(self.updateStackActivityBlock) self.updateStackActivityBlock(_stackActivity);
}

- (void) addedPullActivities: (NSArray *) activities {
	[_stackActivity addObjectsFromArray:activities];
	[_sentStackActivity removeObjectsInArray:activities];
	[self startTimer];
	if(self.updateStackActivityBlock) self.updateStackActivityBlock(_stackActivity);
}

- (void) sendCurrentStackActivities {
	[self sendCurrentStackActivitiesCompletionBlock:nil];
}

- (void) sendCurrentStackActivitiesCompletionBlock: (ObjectBlock) completionBlock {
	if ([_stackActivity count] < 1) {
		if ([self.sendTimer isValid]) {
			[self.sendTimer invalidate];
			self.sendTimer = nil;
		}
		return;
	}
	[self sendHistoryActivityWithUserToken:[VMConsumerManager sharedInstance].userToken
						   completionBlock:^{
							   if (completionBlock) completionBlock(nil);
						   }
								 failBlock:^(NSError *error) {
									 if(completionBlock) completionBlock(error);
								 }];
}

#pragma mark - Background Methods

- (void) willResignActive: (NSNotification *) note {
	if ([self.sendTimer isValid]) {
		[self.sendTimer invalidate];
		self.sendTimer = nil;
	}
	if ([_stackActivity count] > 0) {
		self.backgroundSendID = [[UIApplication sharedApplication] beginBackgroundTaskWithExpirationHandler:^{
			
			NSArray *list = [serviceProvider.operationQueue operations];
			for(VMHTTPRequestOperation *operation in list) {
				[operation cancel];
			}
			
			[_stackActivity writeToFile:[self stackActivityPatch] atomically:YES];
			[_stackActivity removeAllObjects];
			[[UIApplication sharedApplication] endBackgroundTask:self.backgroundSendID];
			
		}];
		[self sendCurrentStackActivities];
	}
}

- (void) didEnterForeground: (NSNotification *) note {
	NSMutableArray *list = [NSMutableArray arrayWithContentsOfFile:[self stackActivityPatch]];
	if (list) {
		[_stackActivity addObjectsFromArray:list];
	}
	[self validateActivitiesLog];
	self.sentStackActivity = [NSMutableArray array];
	if (!_stackActivity) {
		_stackActivity = [NSMutableArray new];
	}
	else {
		[self sendCurrentStackActivities];
	}
	
	//added Notification
	self.backgroundSendID = UIBackgroundTaskInvalid;
}

- (void) endBackgroundTask {
	if ([[UIDevice currentDevice] isMultitaskingSupported] && self.backgroundSendID != UIBackgroundTaskInvalid) {
		[[UIApplication sharedApplication] endBackgroundTask:self.backgroundSendID];
		self.backgroundSendID = UIBackgroundTaskInvalid;
	}
}

#pragma mark - Send Methods

- (VMHTTPRequestOperation *) sendHistoryActivityWithUserToken: (NSString *) userToken
											  completionBlock: (SuccessBlock) completionBlock
													failBlock: (ErrorBlock) errorBlock {
	[VMLog log:[NSString stringWithFormat:@"sendstackActivity send-%@",_stackActivity]];
	VMActivitiesRequest *request = [[VMActivitiesRequest alloc] initWithUserToken:userToken bodyActivities:[self getSendingActivity]];
	VMHTTPRequestOperation *operation = [serviceProvider sendRequest:request success:^(VMActivitiesRequest *request) {
		[_sentStackActivity removeObjectsInArray:request.sentStackActivity];
		if (completionBlock) {
			completionBlock(request.responseDictionary);
		}
		[self sendLastStack];
		[self endBackgroundTask];
		
	} failure:^(VMActivitiesRequest *request, NSError *requestError) {
		[self addedPullActivities:request.sentStackActivity];
		if (errorBlock) {
			errorBlock(requestError);
		}
		[self endBackgroundTask];
	}];
	
	operation.cancelCallback = ^(VMActivitiesRequest *request) {
		[self addedPullActivities:request.sentStackActivity];
	};

	return operation;
}

#pragma mark - Universal

- (void) universalActivityWithActionTypeCode: (VMActionTypeCode) actionTypeCode
								  actionCode: (NSString *) actionCode
									  value1: (NSString *) value1
									  value2: (NSString *) value2
									  value3: (NSString *) value3
								redemptionId: (NSString *) redemptionId
									latitude: (NSString *) latitude
								   longitude: (NSString *) longitude
									 ownerId: (NSString *) ownerId
									   macId: (NSString *) macId
								  merchantId: (NSString *) merchantId
								  campaignId: (NSString *) campaignId
									  itemId: (NSString *) itemId
									itemCode: (NSString *) itemCode
									 venueId: (NSString *) venueId {
	[self addedPullActivityDictionary:[VMActivity mutableDictionaryWithActionTypeCode:actionTypeCode
																		   actionCode:actionCode
																			   value1:value1
																			   value2:value2
																			   value3:value3
																		 redemptionId:redemptionId
																			 latitude:latitude
																			longitude:longitude
																			  ownerId:ownerId
																				macId:macId
																		   merchantId:merchantId
																		   campaignId:campaignId
																			   itemId:itemId
																			 itemCode:itemCode
																			  venueId:venueId]];
}

#pragma mark - Button Click

- (void) buttonClickAnonymousWithButtonId: (NSString *) buttonId
							   merchantId: (NSString *) merchantId
								  venueId: (NSString *) venueId
								   itemId: (NSString *) itemId
								 itemCode: (NSString *) itemCode {
	[self addedPullActivityDictionary:[VMActivity dictionaryButtonClickAnonymousWithButtonId:buttonId
																				  merchantId:merchantId
																					 venueId:venueId
																					  itemId:itemId
																					itemCode:itemCode]];
}

- (void) buttonClickWithButtonId: (NSString *) buttonId
					  merchantId: (NSString *) merchantId
						 venueId: (NSString *) venueId
						  itemId: (NSString *) itemId
						itemCode: (NSString *) itemCode {
	[self addedPullActivityDictionary:[VMActivity dictionaryButtonClickWithButtonId:buttonId
																		 merchantId:merchantId
																			venueId:venueId
																			 itemId:itemId
																		   itemCode:itemCode]];
}

#pragma mark - Location Checkin

- (void) locationCheckinWithTypeGPS: (VMTypeGPS) typeGPS
						   location: (CLLocation *) location {
	[self locationCheckinWithTypeGPS:typeGPS
							location:location
					 completionBlock:nil];
}

- (void) locationCheckinWithTypeGPS: (VMTypeGPS) typeGPS
						   location: (CLLocation *) location
					completionBlock: (ObjectBlock) completionBlock {
	NSString *accuracy = [NSString stringWithFormat:@"%f:%f", location.horizontalAccuracy, location.verticalAccuracy];
	[self locationCheckinWithTypeGPS:typeGPS
					locationAccuracy:accuracy
					   currentCourse:[NSString stringWithFormat:@"%f",location.course]
							   speed:[NSString stringWithFormat:@"%f",location.speed]
							altitude:[NSString stringWithFormat:@"%f",location.altitude]
							latitude:[NSString stringWithFormat:@"%f",location.coordinate.latitude]
						   longitude:[NSString stringWithFormat:@"%f",location.coordinate.longitude]
					 completionBlock:completionBlock];
}

- (void) locationCheckinWithTypeGPS: (VMTypeGPS) typeGPS
				   locationAccuracy: (NSString *) locationAccuracy
					  currentCourse: (NSString *) currentCourse
							  speed: (NSString *) speed
						   altitude: (NSString *) altitude
						   latitude: (NSString *) latitude
						  longitude: (NSString *) longitude {
	[self locationCheckinWithTypeGPS:typeGPS
					locationAccuracy:locationAccuracy
					   currentCourse:currentCourse
							   speed:speed
							altitude:altitude
							latitude:latitude
						   longitude:longitude
					 completionBlock:nil];
}

- (void) locationCheckinWithTypeGPS: (VMTypeGPS) typeGPS
				   locationAccuracy: (NSString *) locationAccuracy
					  currentCourse: (NSString *) currentCourse
							  speed: (NSString *) speed
						   altitude: (NSString *) altitude
						   latitude: (NSString *) latitude
						  longitude: (NSString *) longitude
					completionBlock: (ObjectBlock) completionBlock {
	[self addedPullActivityDictionary:[VMActivity dictionaryLocationCheckinWithTypeGPS:typeGPS
																	  locationAccuracy:locationAccuracy
																		 currentCourse:currentCourse
																				 speed:speed
																			  altitude:altitude
																			  latitude:latitude
																			 longitude:longitude]];
	[self sendCurrentStackActivitiesCompletionBlock:completionBlock];
}

#pragma mark - AppPageImpression Methods

- (void) appPageImpressionAnonymWithPageId: (NSString *) pageId
								merchantId: (NSString *) merchantId
								   venueId: (NSString *) venueId
									itemId: (NSString *) itemId
								  itemCode: (NSString *) itemCode {
	[self addedPullActivityDictionary:[VMActivity dictionaryAppPageImpressionAnonymWithPageId:pageId
																				   merchantId:merchantId
																					  venueId:venueId
																					   itemId:itemId
																					 itemCode:itemCode]];
}

- (void) appPageImpressionWithPageId: (NSString *) pageId
						  merchantId: (NSString *) merchantId
							 venueId: (NSString *) venueId
							  itemId: (NSString *) itemId
							itemCode: (NSString *) itemCode {
	[self addedPullActivityDictionary:[VMActivity dictionaryAppPageImpressionWithPageId:pageId
																			 merchantId:merchantId
																				venueId:venueId
																				 itemId:itemId
																			   itemCode:itemCode]];
}

#pragma mark - Offer

- (void) offerImpressionWithPlacementType: (NSString *) placementType
								  offerId: (NSString *) offerId {
	[self addedPullActivityDictionary:[VMActivity dictionaryOfferImpressionWithPlacementType:placementType
																					 offerId:offerId]];
}

- (void) offerRedemptionWithPlacementType: (NSString *) placementType
						  paymentTypeCode: (NSString *) paymentTypeCode
								paymentId: (NSString *) paymentId
							paymentAmount: (NSString *) paymentAmount
							redeemedOffer: (NSString *) redeemedOffer
								  offerId: (NSString *) offerId
								  venueId: (NSString *) venueId {
	[self addedPullActivityDictionary:[VMActivity dictionaryOfferRedemptionWithPlacementType:placementType
																			 paymentTypeCode:paymentTypeCode
																				   paymentId:paymentId
																			   paymentAmount:paymentAmount
																			   redeemedOffer:redeemedOffer
																					 offerId:offerId
																					 venueId:venueId]];
}

- (void) offerClickThroughWithPlacementType: (NSString *) placementType
									offerId: (NSString *) offerId {
	[self addedPullActivityDictionary:[VMActivity dictionaryOfferClickThroughWithPlacementType:placementType
																					   offerId:offerId]];
}

- (void) offerInstoreRedemptionWithPlacementType: (NSString *) placementType
								   redeemedOffer: (NSString *) redeemedOffer
										 offerId: (NSString *) offerId
										 venueId: (NSString *) venueId {
	[self addedPullActivityDictionary:[VMActivity dictionaryOfferInstoreRedemptionWithPlacementType:placementType
																					  redeemedOffer:redeemedOffer
																							offerId:offerId
																							venueId:venueId]];
}

#pragma mark - Application

- (void) appStartup {
	[self addedPullActivityDictionary:[VMActivity dictionaryAppStartup]];
}

- (void) appStartupAnonym {
	[self addedPullActivityDictionary:[VMActivity dictionaryAppStartupAnonym]];
}

- (void) appInstall {
	[self addedPullActivityDictionary:[VMActivity dictionaryAppInstall]];
}

#pragma mark - Social

- (void) registrationWithRegistrationType: (VMActivitySocialType) socialType {
	[self addedPullActivityDictionary:[VMActivity dictionaryRegistrationWithRegistrationType:socialType]];
}

- (void) loginWithLoginType: (VMActivitySocialType) socialType {
	[self addedPullActivityDictionary:[VMActivity dictionaryLoginWithLoginType:socialType]];
}

- (void) offerShareWithSocialType: (VMShareSocialType) socialType
						  offerId: (NSString *) offerId {
	[self addedPullActivityDictionary:[VMActivity dictionaryOfferShareWithSocialType:socialType
																			 offerId:offerId]];
}

#pragma mark - Advertisement

- (void) advertisementImpressionWithPlacementCode: (NSString *) placementCode
									  channelCode: (NSString *) channelCode
								  advertisementId: (NSString *) advertisementId {
	[self addedPullActivityDictionary:[VMActivity dictionaryAdvertisementImpressionWithPlacementCode:placementCode
																						 channelCode:channelCode
																					 advertisementId:advertisementId]];
}

- (void) advertisementClickWithPlacementCode: (NSString *) placementCode
								 channelCode: (NSString *) channelCode
							 advertisementId: (NSString *) advertisementId {
	[self addedPullActivityDictionary:[VMActivity dictionaryAdvertisementClickWithPlacementCode:placementCode
																					channelCode:channelCode
																				advertisementId:advertisementId]];
}

#pragma mark - 

- (void)dealloc {
	[_stackActivity addObjectsFromArray:_sentStackActivity];
	[_stackActivity writeToFile:[self stackActivityPatch] atomically:YES];
	[[NSNotificationCenter defaultCenter] removeObserver:self
													name:UIApplicationWillResignActiveNotification
												  object:nil];
	[[NSNotificationCenter defaultCenter] removeObserver:self
													name:UIApplicationWillEnterForegroundNotification
												  object:nil];
}

@end
