//
//  VMGEOFenceLocationManager.m
//  geofencedemo
//
//  Created by Ali Darvish on 12/09/13.
//  Copyright (c) 2013 Kevin McMahon. All rights reserved.
//

#import "VMGEOFenceLocationManager.h"
#import "VMUtils.h"
#import "VMActivityManager.h"
#import "VMLog.h"
#import "VMGeoFencesRequest.h"
#import "VMConsumerManager.h"

#define kVMMorningTime 900
#define kVMNightTime 2100
#define kRequiredHorizontalAccuracy 20.0

@interface VMGEOFenceLocationManager ()
@property (nonatomic, strong) CLLocation *currLocation;
@property (nonatomic, copy) ObjectBlock updateActivityLocationBlock;
@property (nonatomic, strong) NSMutableArray *currGeo;
@end

@implementation VMGEOFenceLocationManager

@synthesize locationManager, updateLocationTimer, thread1, IsSystemSleeping;

+ (instancetype)sharedInstance;
{
    static dispatch_once_t once;
    static id sharedInstance;
    dispatch_once(&once, ^{
        sharedInstance = [[self alloc] init];
    });
    return sharedInstance;
}

//start location services
- (void) startLocationService
{
    [self initLocationManagerWithRegions];
}

//stop location services
- (void) stopLocationServices
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    [self.updateLocationTimer invalidate];
    [self.locationManager stopMonitoringSignificantLocationChanges];
    [self.locationManager stopUpdatingLocation];
    self.locationManager = nil;
}

- (void) logOnTextView:(UITextView *)txt
{
    data = txt;
}

//resume location service update
- (void) startSystem
{
    if(self.IsSystemSleeping)
    {
        [self.updateLocationTimer setFireDate:[NSDate dateWithTimeIntervalSinceNow:60]];
        [self.locationManager stopMonitoringSignificantLocationChanges];
        [self.locationManager startUpdatingLocation];
        self.IsSystemSleeping = NO;
    }
}

//stop location service update momentary
- (void) stopSystem
{
    [self.locationManager stopUpdatingLocation];
    [self.locationManager startMonitoringSignificantLocationChanges];
    self.IsSystemSleeping = YES;
    [self.updateLocationTimer setFireDate:[NSDate distantFuture]];
}

//--------------------------------------------------------------------------------------------------------------
//timer thread set up to update user location with server every one hour (will run in the background)
- (void) initUpdateTimer:(NSThread *)th
{
    [self.updateLocationTimer invalidate];
    self.updateLocationTimer = nil;
    self.updateLocationTimer = [NSTimer timerWithTimeInterval:60 target:self selector:@selector(updateRegions:) userInfo:nil repeats:YES];
    [[NSRunLoop mainRunLoop] addTimer:self.updateLocationTimer forMode:NSRunLoopCommonModes];
}

//the timer will stop region monitoring before 9am and after 9pm
- (void) updateRegions:(NSTimer *)timer
{
    if((kVMNightTime <= [self getHourFromDate]) || (kVMMorningTime >= [self getHourFromDate]))
    {
        [self stopSystem];
        NSLog(@"Stop Everything Temp");
        data.text = [data.text stringByAppendingString:@"Stop Everything Temp"];
    }
    else
    {
        NSLog(@"Make Request");
        [self updateRegions];
    }
}
//--------------------------------------------------------------------------------------------------------------

- (void) updateRegions
{
    NSString *lat = [[NSNumber numberWithDouble:self.currLocation.coordinate.latitude] stringValue];
    NSString *longt = [[NSNumber numberWithDouble:self.currLocation.coordinate.longitude] stringValue];
    
    [self getGeoFencesWithUserToken:[VMConsumerManager sharedInstance].userToken latitude:lat longitude:longt completionBlock:^(NSArray *coordinates){
        NSLog(@"%@", coordinates);
        
        if(coordinates.count != 0)
            [self initializeRegionMonitoring:coordinates];
        else
            [self stopSystem];
    } failBlock:^(NSError *error){
        NSLog(@"%@", error.debugDescription);
        [self stopSystem];
    }];
}

//--------------------------------------------------------------------------------------------------------------
//init location manager
- (void) initLocationManager
{
    if(![CLLocationManager locationServicesEnabled])
    {
        [[[UIAlertView alloc] initWithTitle:@"Location Service" message:@"You need to enable location services to use this app." delegate:nil cancelButtonTitle:@"Ok" otherButtonTitles:nil, nil] show];
        
        return;
    }
    
    if(![CLLocationManager regionMonitoringAvailable])
    {
        [[[UIAlertView alloc] initWithTitle:@"Location Service" message:@"This app requires region monitoring features which are unavailable on this device." delegate:nil cancelButtonTitle:@"Ok" otherButtonTitles:nil, nil] show];
        
        return;
    }
    
    self.locationManager = [[CLLocationManager alloc] init];
    self.locationManager.desiredAccuracy = kCLLocationAccuracyBest;
    self.locationManager.distanceFilter = kCLDistanceFilterNone;
    self.locationManager.delegate = self;
}
//--------------------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------------------
//add geofence data(region object) to location manager for monitoring
- (void) initializeRegionMonitoring:(NSArray *)geofences
{
    for(NSDictionary *dictionary in geofences)
    {
        CLLocation *location = [[CLLocation alloc] initWithLatitude:[[dictionary valueForKey:@"latitude"] doubleValue] longitude:[[dictionary valueForKey:@"longitude"] doubleValue]];
        if([self isValidLocation:location])
        {
            CLLocationDegrees latitude = [[dictionary valueForKey:@"latitude"] doubleValue];
            CLLocationDegrees longitude = [[dictionary valueForKey:@"longitude"] doubleValue];
            CLLocationCoordinate2D centerCoordinate = CLLocationCoordinate2DMake(latitude, longitude);
    
            CLLocationDistance regionRadius = [[dictionary valueForKey:@"radius"] doubleValue];
    
            CLRegion *region = [[CLRegion alloc] initCircularRegionWithCenter:centerCoordinate
                                                   radius:regionRadius
                                               identifier:[[dictionary valueForKey:@"latitude"] stringValue]];
    
   
            [self.locationManager startMonitoringForRegion:region];
        }
    }
}

- (void) initLocationManagerWithRegions
{
    self.IsSystemSleeping = NO;
    serviceProvider = nil;
    serviceProvider = [[VMServiceProvider alloc] initWithBaseURL:[NSURL URLWithString:[VMUtils offerApiUrl]]];
    
    if(self.locationManager == nil)
    {
        [self initLocationManager];
        [self.locationManager startUpdatingLocation];
    }
    
    if(self.locationManager != nil)
    {
        if(self.thread1 == nil)
        {
            [NSTimer scheduledTimerWithTimeInterval:5 target:self selector:@selector(startLoadingDataOnInit:) userInfo:nil repeats:NO];
            
            [[NSNotificationCenter defaultCenter] removeObserver:self];
            [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(appBecomeActive:) name:UIApplicationWillEnterForegroundNotification object:nil];
            
            self.thread1 = nil;
            self.thread1 = [[NSThread alloc] initWithTarget:self selector:@selector(initUpdateTimer:) object:nil];
            [self.thread1 start];
        }
    }
}

- (void) startLoadingDataOnInit:(NSTimer *)timer
{
    [self updateRegions];
}
//--------------------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------------------
//Region Request
- (VMHTTPRequestOperation *) getGeoFencesWithUserToken:(NSString *)token latitude:(NSString *)latitude longitude:(NSString *)longitude completionBlock: (ObjectBlock) completionBlock failBlock: (ErrorBlock) errorBlock
{
    VMGeoFencesRequest *request = [[VMGeoFencesRequest alloc] initWithUserToken:token latitude:latitude longitude:longitude];
    
    return [serviceProvider sendRequest:request success:^(VMGeoFencesRequest *request){
        if(completionBlock)
            completionBlock(request.geofences);
    } failure:^(VMBaseRequest *req, NSError *error){
        if(errorBlock)
            errorBlock(error);
    }];
}
//--------------------------------------------------------------------------------------------------------------

- (NSInteger) getHourFromDate
{
    NSDateFormatter *df = [[NSDateFormatter alloc] init];
    [df setDateFormat:@"HHmm"];
    NSString *hourStr = [df stringFromDate:[NSDate date]];
    
    return [hourStr integerValue];
}

//--------------------------------------------------------------------------------------------------------------
//send user location to server
- (void) sendActivityManager
{
	UIApplicationState state = [UIApplication sharedApplication].applicationState;
	[VMLog log:[NSString stringWithFormat:@"UIApplicationState %d",state]];
	
		
    [[VMActivityManager sharedInstance] locationCheckinWithTypeGPS:VMGPS location:self.currLocation completionBlock:^(id description) {
			[VMLog log:[NSString stringWithFormat:@"battery Level -%f ",[UIDevice currentDevice].batteryLevel]];
    }];
    
    if (self.updateActivityLocationBlock) self.updateActivityLocationBlock(self.currLocation);
}
//--------------------------------------------------------------------------------------------------------------

- (BOOL) isValidLocation:(CLLocation *)newLocation
{
	// not sending (0,0) coordinates because we have found out that sometimes the devices that have GPS but no SIM-card inserted return (0,0) coordinated. Most probably, no one ever will be using the app in the (0,0) point
	if(newLocation.horizontalAccuracy == 0 && newLocation.coordinate.longitude == 0 && newLocation.coordinate.latitude == 0)
    {
		return NO;
	}
	else
    {
		return YES;
	}
}

//--------------------------------------------------------------------------------------------------------------
#pragma mark - Location Delegate

- (void) locationManager:(CLLocationManager *)manager didEnterRegion:(CLRegion *)region
{
    NSLog(@"Did enter region: %f, %f", region.center.latitude, region.center.longitude);
    data.text = [data.text stringByAppendingFormat:@"Did enter region: %f, %f\n\n", region.center.latitude, region.center.longitude];
    [self sendActivityManager];
    [locationManager stopMonitoringForRegion:region];
}

- (void) locationManager:(CLLocationManager *)manager didStartMonitoringForRegion:(CLRegion *)region
{
    NSLog(@"Did Start monitoring for region: %f, %f", region.center.latitude, region.center.longitude);
    data.text = [data.text stringByAppendingFormat:@"Did Start monitoring for region: %f, %f\n\n", region.center.latitude, region.center.longitude];
}

- (void) locationManager:(CLLocationManager *)manager monitoringDidFailForRegion:(CLRegion *)region withError:(NSError *)error
{
    
}

- (void) locationManager:(CLLocationManager *)manager didUpdateToLocation:(CLLocation *)newLocation fromLocation:(CLLocation *)oldLocation
{
    if(newLocation.horizontalAccuracy <= kRequiredHorizontalAccuracy)
    {
        self.locationManager.desiredAccuracy = kCLLocationAccuracyBest;
    }
    else
    {
        self.locationManager.desiredAccuracy = kCLLocationAccuracyThreeKilometers;
    }

    self.currLocation = newLocation;
    [VMUtils setCLLocation:newLocation];
    [self startSystem];
}
//--------------------------------------------------------------------------------------------------------------

- (void) appBecomeActive:(NSNotification *)notf
{
    [self initLocationManagerWithRegions];
    [self startSystem];
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:UIApplicationWillEnterForegroundNotification object:nil];
}

@end
