//
//  VMLocationsManager.m
//  VMobSDK
//
//  Created by vmob.co on 6/3/13.
//  Copyright (c) 2013 vmob.co. All rights reserved.
//

#import "VMLocationsManager.h"
#import "VMUtils.h"
#import "VMStatesRequest.h"
#import "VMCitiesRequest.h"
#import "VMActivityManager.h"
#import "VMLog.h"

@interface VMLocationsManager()<CLLocationManagerDelegate>
@property (nonatomic, strong) CLLocationManager *locationManager;
@property (nonatomic, strong) CLLocationManager *backgroundManager;
@property (nonatomic, strong) NSTimer *updateLocationTimer;
@property (nonatomic, strong) NSDate *lastSendDate;
@property (nonatomic, copy) ObjectBlock updateBlock;
@property (nonatomic, copy) ErrorBlock errorBlock;
@property (nonatomic, copy) ObjectBlock updateActivityLocationBlock;
@property (nonatomic) UIBackgroundTaskIdentifier backgroundSendID;
@property (nonatomic) BOOL isUpdateCurrentLocation;
@property (nonatomic) BOOL isUpdating;

@end

@implementation VMLocationsManager

- (id)init {
	if (self = [super init]) {
		serviceProvider = [[VMServiceProvider alloc] initWithBaseURL:[NSURL URLWithString:[VMUtils locationApiUrl]]];
		[self createLocationManager:UIApplicationStateActive];
		[self startLocation];
		[[NSNotificationCenter defaultCenter] addObserver:self
												 selector:@selector(willResignActive:)
													 name:UIApplicationWillResignActiveNotification
												   object:nil];
		[[NSNotificationCenter defaultCenter] addObserver:self
												 selector:@selector(didEnterForeground:)
													 name:UIApplicationWillEnterForegroundNotification
												   object:nil];
	}
	return self;
}

- (void) createLocationManager: (UIApplicationState) applicationState {
	if (applicationState == UIApplicationStateActive) {
		[self.backgroundManager stopMonitoringSignificantLocationChanges];
		self.backgroundManager.delegate = nil;
		self.backgroundManager = nil;
		
		self.locationManager = [[CLLocationManager alloc] init];
		self.locationManager.desiredAccuracy = [VMUtils locationAccuracy];
		self.locationManager.distanceFilter = [VMUtils locationDistanceFilter];
		self.locationManager.delegate = self;
		[self startLocationWithCompletionBlock:self.updateBlock failBlock:self.errorBlock];
	}
	else {
		if (self.isUpdating) {
			[self.locationManager stopUpdatingLocation];
			self.locationManager.delegate = nil;
			self.locationManager = nil;
			
			self.backgroundManager = [[CLLocationManager alloc] init];
			self.backgroundManager.delegate = self;
			[self.backgroundManager startMonitoringSignificantLocationChanges];
		}
	}
}

+ (instancetype) sharedInstance {
	static dispatch_once_t once;
    static id sharedInstance;
    dispatch_once(&once, ^{
        sharedInstance = [[self alloc] init];
    });
    return sharedInstance;
}

#pragma mark - Public Methods

- (VMHTTPRequestOperation *)statesListWithCompletionBlock: (ObjectBlock) completionBlock
												failBlock: (ErrorBlock) errorBlock {
	VMStatesRequest *request = [[VMStatesRequest alloc] init];
	return [serviceProvider sendRequest:request success:^(VMStatesRequest *request) {
		if (completionBlock) {
			completionBlock(request.statesList);
		}
	} failure:^(VMBaseRequest * request, NSError *requestError) {
		if (errorBlock) {
			errorBlock(requestError);
		}
	}];
}

- (VMHTTPRequestOperation *)citiesListWithCompletionBlock: (ObjectBlock) completionBlock
												failBlock: (ErrorBlock) errorBlock {
	VMCitiesRequest *request = [[VMCitiesRequest alloc] init];
	return [serviceProvider sendRequest:request success:^(VMCitiesRequest *request) {
		if (completionBlock) {
			completionBlock(request.citiesList);
		}
	} failure:^(VMBaseRequest * request, NSError *requestError) {
		if (errorBlock) {
			errorBlock(requestError);
		}
	}];
}

#pragma mark - Private Methods

- (void) sendActivityManager {
	UIApplicationState state = [UIApplication sharedApplication].applicationState;
	[VMLog log:[NSString stringWithFormat:@"UIApplicationState %d",state]];
	if (_currentLocation && self.isUpdateCurrentLocation) {
		self.isUpdateCurrentLocation = NO;
		[[VMActivityManager sharedInstance] locationCheckinWithTypeGPS:VMGPS location:_currentLocation completionBlock:^(id description) {
			[VMLog log:[NSString stringWithFormat:@"battery Level -%f ",[UIDevice currentDevice].batteryLevel]];
		}];
		if (self.updateActivityLocationBlock) self.updateActivityLocationBlock(_currentLocation);
	}
}

#pragma mark - Follow Methods

- (void) setUpdateActivityLocation: (ObjectBlock) updateActivityLocationBlock {
	self.updateActivityLocationBlock = updateActivityLocationBlock;
}

- (void) setUpdateErrorBlock: (ErrorBlock) errorBlock {
	self.errorBlock = errorBlock;
}

- (void) setUpdateLocation: (ObjectBlock) block {
	self.updateBlock = block;
}

- (BOOL) isUpdatingLocation {
	return self.isUpdating;
}

- (NSTimeInterval) updateLocationTimeInterval {
	return [VMUtils updateLocationTimeInterval];
}

#pragma mark - Check Location

- (void) stopLocation {
	self.isUpdating = NO;
	[self.locationManager stopUpdatingLocation];
	if ([self.updateLocationTimer isValid]) {
		[self.updateLocationTimer invalidate];
		self.updateLocationTimer = nil;
	}
}

- (void) startLocation {
	[self startLocationWithCompletionBlock:self.updateBlock failBlock:self.errorBlock];
}

- (void) startLocationWithCompletionBlock: (ObjectBlock) updateBlock
								failBlock: (ErrorBlock) errorBlock {
	self.updateBlock = updateBlock;
	self.errorBlock = errorBlock;
	self.isUpdating = YES;
	[self.locationManager startUpdatingLocation];
	
	if (self.updateLocationTimer) {
		if ([self.updateLocationTimer isValid]) {
			[self.updateLocationTimer invalidate];
		}
		self.updateLocationTimer = nil;
	}
	self.updateLocationTimer = [NSTimer scheduledTimerWithTimeInterval:[VMUtils updateLocationTimeInterval] target:self selector:@selector(sendActivityManager) userInfo:nil repeats:YES];
}

#pragma mark - Background Methods

- (void) willResignActive: (NSNotification *) note {
	[self createLocationManager:UIApplicationStateBackground];
}

- (void) didEnterForeground: (NSNotification *) note {
	[self createLocationManager:UIApplicationStateActive];
}

- (BOOL) checkTimer {
	if (self.lastSendDate) {
		NSTimeInterval interval = [[NSDate date] timeIntervalSinceDate:self.lastSendDate];
		return (interval > [VMUtils gpsBackgroundMinIntervalInSeconds]);
	}
	return YES;
}

- (void) sendBackgroundMetod: (CLLocation *) newLocation {
	UIApplicationState state = [UIApplication sharedApplication].applicationState;
	[VMLog log:[NSString stringWithFormat:@"UIApplicationState %d",state]];
	if ([UIApplication sharedApplication].applicationState == UIApplicationStateBackground && [self checkTimer] && newLocation) {
		
		UIBackgroundTaskIdentifier backgroundSendID = [[UIApplication sharedApplication] beginBackgroundTaskWithExpirationHandler:^{
			
		}];
		self.lastSendDate = [NSDate date];
		[[VMActivityManager sharedInstance] locationCheckinWithTypeGPS:VMGPS location:newLocation completionBlock:^(id description) {
			self.lastSendDate = [NSDate date];
			[VMLog log:[NSString stringWithFormat:@"battery Level -%f ",[UIDevice currentDevice].batteryLevel]];
			[[UIApplication sharedApplication] endBackgroundTask:backgroundSendID];
		}];
	}
}

- (BOOL) isValidLocation:(CLLocation *)newLocation {
	// not sending (0,0) coordinates because we have found out that sometimes the devices that have GPS but no SIM-card inserted return (0,0) coordinated. Most probably, no one ever will be using the app in the (0,0) point
	if(newLocation.horizontalAccuracy == 0 && newLocation.coordinate.longitude == 0 && newLocation.coordinate.latitude == 0) {
		return NO;
	}
	else {
		return YES;
	}
}

#pragma mark - CLLocationManagerDelegate

- (void)locationManager: (CLLocationManager *) manager monitoringDidFailForRegion:(CLRegion *)region withError:(NSError *)error {
	[VMLog log:[NSString stringWithFormat:@"Error location - %@", error]];
	if (self.errorBlock) self.errorBlock(error);
}

- (void)locationManager:(CLLocationManager *)manager didUpdateToLocation:(CLLocation *)newLocation fromLocation:(CLLocation *)oldLocation {
	[VMLog log:[NSString stringWithFormat:@"didUpdateToLocation %@", newLocation]];
	if (![self isValidLocation:newLocation]) {
		return;
	}
	[VMUtils setCLLocation:newLocation];
	[self sendBackgroundMetod:newLocation];
	if (newLocation.horizontalAccuracy <= [VMUtils locationAccuracy] || !self.currentLocation) {
		self.isUpdateCurrentLocation = YES;
		
		if (!self.currentLocation) {
			_currentLocation = newLocation;
			[self sendActivityManager];
		}
		_currentLocation = newLocation;
		if(self.updateBlock) self.updateBlock(_currentLocation);
	}
}

#pragma mark - 

- (void)dealloc {
	[[NSNotificationCenter defaultCenter] removeObserver:self
													name:UIApplicationWillResignActiveNotification
												  object:nil];
	[[NSNotificationCenter defaultCenter] removeObserver:self
													name:UIApplicationWillEnterForegroundNotification
												  object:nil];
}

@end
